/*
 * Copyright (C) 2011-2012 jeanfi@gmail.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */

#include <libintl.h>
#define _(String) gettext(String)

#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <curl/curl.h>
#include <json/json.h>

#include "cache.h"
#include "list.h"
#include "log.h"
#include "lp_ws.h"
#include "lp_json.h"
#include "ppastats.h"

static const char *
QUERY_GET_PUBLISHED_BINARIES = "?ws.op=getPublishedBinaries&ws.size=300";
static const char *QUERY_GET_DOWNLOAD_COUNT = "?ws.op=getDownloadCount";
static const char *
QUERY_GET_DAILY_DOWNLOAD_TOTALS = "?ws.op=getDailyDownloadTotals";

static const int DEFAULT_FETCH_RETRIES = 10;

static CURL *curl;

struct ucontent {
	char *data;
	size_t len;
};

static size_t cbk_curl(void *buffer, size_t size, size_t nmemb, void *userp)
{
	size_t realsize = size * nmemb;
	struct ucontent *mem = (struct ucontent *)userp;

	mem->data = realloc(mem->data, mem->len + realsize + 1);

	memcpy(&(mem->data[mem->len]), buffer, realsize);
	mem->len += realsize;
	mem->data[mem->len] = 0;

	return realsize;
}

static void init()
{
	if (!curl) {
		log_debug(_("initializing CURL"));
		curl_global_init(CURL_GLOBAL_ALL);
		curl = curl_easy_init();
	}

	if (!curl)
		exit(EXIT_FAILURE);
}

static char *fetch_url(const char *url)
{
	struct ucontent *content = malloc(sizeof(struct ucontent));
	char *result;
	long code;
	int retries;
	unsigned int s;

	log_debug(_("fetch_url(): %s"), url);

	init();

	result = NULL;

	retries = DEFAULT_FETCH_RETRIES;

 retrieve:
	content->data = malloc(1);
	content->data[0] = '\0';
	content->len = 0;

	curl_easy_setopt(curl, CURLOPT_URL, url);
	curl_easy_setopt(curl, CURLOPT_VERBOSE, 0);
	curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, cbk_curl);
	curl_easy_setopt(curl, CURLOPT_WRITEDATA, content);
	curl_easy_setopt(curl, CURLOPT_USERAGENT, "ppastats/0.0");

	if (curl_easy_perform(curl) == CURLE_OK) {
		curl_easy_getinfo(curl, CURLINFO_RESPONSE_CODE, &code);

		switch (code) {
		case 200:
			result = content->data;
			break;
		case 500:
		case 502:
		case 503:
		case 504:
			log_err(_("Fetch failed with code %ld for URL %s"),
				code,
				url);

			if (retries) {
				s = 2 * (DEFAULT_FETCH_RETRIES - retries) + 2;
				log_debug(_("Wait %ds before retry"), s);
				sleep(s);

				free(content->data);
				retries--;
				goto retrieve;
			}

			break;
		default:
			log_err(_("Fetch failed with code %ld for URL %s"),
				code,
				url);
		}
	}

	if (!result)
		free(content->data);

	free(content);

	return result;
}

static json_object *get_json_object(const char *url)
{
	json_object *obj = NULL;
	char *body;

	body = fetch_url(url);

	if (body) {
		obj = json_tokener_parse(body);

		free(body);

		return obj;
	}

	return NULL;
}

#define json_object_to_bpph_list \
json_object_to_binary_package_publishing_history_list

struct binary_package_publishing_history * *
get_binary_package_publishing_history_list(const char *archive_url,
					   const char *pkg_status)
{
	struct json_object *o_next;
	char *url;
	json_object *o;
	void **result = NULL;

	url = malloc(strlen(archive_url)+
		     strlen(QUERY_GET_PUBLISHED_BINARIES)+
		     (pkg_status ? strlen("&status=")+strlen(pkg_status) : 0)+
		     1);

	strcpy(url, archive_url);
	strcat(url, QUERY_GET_PUBLISHED_BINARIES);

	if (pkg_status) {
		strcat(url, "&status=");
		strcat(url, pkg_status);
	}

	while (url) {
		o = get_json_object(url);
		free(url);
		url = NULL;

		if (!o)
			break;

		result = list_append_list(result,
					  (void **)json_object_to_bpph_list(o));

		o_next = json_object_object_get(o, "next_collection_link");

		if (o_next)
			url = strdup(json_object_get_string(o_next));

		json_object_put(o);
	}

	return (struct binary_package_publishing_history **)result;
}

int get_download_count(const char *archive_url)
{
	int n = strlen(archive_url) + strlen(QUERY_GET_DOWNLOAD_COUNT) + 1;
	char *url = malloc(n);
	int result;
	json_object *obj;

	strcpy(url, archive_url);
	strcat(url, QUERY_GET_DOWNLOAD_COUNT);

	obj = get_json_object(url);
	free(url);

	if (!obj)
		return -1;

	result = json_object_get_int(obj);

	json_object_put(obj);

	return result;
}

const struct distro_arch_series *get_distro_arch_series(const char *url)
{
	json_object *obj;
	const struct distro_arch_series *distro;

	distro = cache_get(url);
	if (distro)
		return (struct distro_arch_series *)distro;

	obj = get_json_object(url);

	if (!obj)
		return NULL;

	distro = json_object_to_distro_arch_series(obj);

	json_object_put(obj);

	cache_put(url, distro, (void (*)(void *))&distro_arch_series_free);

	return distro;
}

const struct distro_series *get_distro_series(const char *url)
{
	json_object *obj;
	const struct distro_series *distro;

	distro = cache_get(url);
	if (distro)
		return (struct distro_series *)distro;

	obj = get_json_object(url);

	if (!obj)
		return NULL;

	distro = json_object_to_distro_series(obj);

	json_object_put(obj);

	cache_put(url, distro, (void (*)(void *))&distro_series_free);

	return distro;
}

struct daily_download_total **get_daily_download_totals(const char *binary_url)
{
	char *url;
	json_object *obj;
	struct daily_download_total **result = NULL;

	url = malloc(strlen(binary_url)+
		     strlen(QUERY_GET_DAILY_DOWNLOAD_TOTALS)+1);

	strcpy(url, binary_url);
	strcat(url, QUERY_GET_DAILY_DOWNLOAD_TOTALS);

	obj = get_json_object(url);

	if (obj) {
		result = json_object_to_daily_download_totals(obj);
		json_object_put(obj);
	}

	free(url);

	return result;
}

void lp_ws_cleanup()
{
	log_debug(_("cleanup CURL"));

	curl_easy_cleanup(curl);
	curl_global_cleanup();
}
